// AUTO-GENERATED BY UpdateGenVersion.bat ff6621cc5394b46125e5fa453dc91e45addac692
const char GenVersion_Version[] = "0.4.4-dev";
const char GenVersion_Commit[] = "ff6621cc5394b46125e5fa453dc91e45addac692";
